@@ -15,9 +15,9 @@ from requests.exceptions import JSONDecodeError
 from source_shopify.utils import ApiTypeEnum
 from source_shopify.utils import ShopifyRateLimiter as limiter
 
-from ...http_request import ShopifyErrorHandler
 from .exceptions import AirbyteTracedException, ShopifyBulkExceptions
 from .query import ShopifyBulkQuery, ShopifyBulkTemplates
+from .record import ShopifyBulkRecord
 from .retry import bulk_retry_on_exception
 from .status import ShopifyBulkJobStatus
 from .tools import END_OF_FILE, BulkTools
@@ -25,12 +25,12 @@ from .tools import END_OF_FILE, BulkTools
 
 @dataclass
 class ShopifyBulkManager:
-    session: requests.Session
+    http_client: HttpClient
     base_url: str
-    stream_name: str
     query: ShopifyBulkQuery
     job_termination_threshold: float
     job_size: float
+    job_checkpoint_interval: int
 
     # default logger
     logger: Final[logging.Logger] = logging.getLogger("airbyte")
@@ -66,8 +66,11 @@ class ShopifyBulkManager:
 
     # 0.1 ~= P2H, default value, lower boundary for slice size
     _job_size_min: Final[float] = 0.1
-    # P365D, upper boundary for slice size
-    _job_size_max: Final[float] = 365.0
+
+    # last running job object count
+    _job_last_rec_count: int = field(init=False, default=0)
+    # the flag to adjust the next slice from the checkpointed cursor vaue
+    _job_adjust_slice_from_checkpoint: bool = field(init=False, default=False)
 
     # expand slice factor
     _job_size_expand_factor: int = field(init=False, default=2)
@@ -79,14 +82,19 @@ class ShopifyBulkManager:
     # 2 sec is set as default value to cover the case with the empty-fast-completed jobs
     _job_last_elapsed_time: float = field(init=False, default=2.0)
 
-    def __post_init__(self):
-        self._http_client = HttpClient(self.stream_name, self.logger, ShopifyErrorHandler(), session=self.session)
+    def __post_init__(self) -> None:
         self._job_size = self.job_size
+        # The upper boundary for slice size is limited by the value from the config, default value is `P30D`
+        self._job_size_max = self.job_size
         # Each job ideally should be executed within the specified time (in sec),
         # to maximize the performance for multi-connection syncs and control the bulk job size within +- 1 hours (3600 sec),
         # Ideally the source will balance on it's own rate, based on the time taken to return the data for the slice.
         # This behaviour could be overidden by providing the `BULK Job termination threshold` option in the `config`.
         self._job_max_elapsed_time = self.job_termination_threshold
+        # how many records should be collected before we use the checkpoining
+        self._job_checkpoint_interval = self.job_checkpoint_interval
+        # define Record Producer instance
+        self.record_producer: ShopifyBulkRecord = ShopifyBulkRecord(self.query)
 
     @property
     def _tools(self) -> BulkTools:
@@ -143,6 +151,21 @@ class ShopifyBulkManager:
         self._job_should_revert_slice = False
         return False
 
+    @property
+    def _supports_checkpointing(self) -> bool:
+        """
+        The flag to determine whether or not the BULK Stream supports the `BULK checkpointing`.
+        """
+        return self.query.supports_checkpointing
+
+    @property
+    def _job_should_checkpoint(self) -> bool:
+        return self._supports_checkpointing and self._job_last_rec_count >= self._job_checkpoint_interval
+
+    @property
+    def _job_any_lines_collected(self) -> bool:
+        return self._job_last_rec_count > 0
+
     def _expand_job_size(self) -> None:
         self._job_size += self._job_size_adjusted_expand_factor
 
@@ -176,6 +199,16 @@ class ShopifyBulkManager:
         self._job_self_canceled = False
         # set the running job message counter to default
         self._log_job_msg_count = 0
+        # set the running job object count to default
+        self._job_last_rec_count = 0
+
+    def _set_checkpointing(self) -> None:
+        # set the flag to adjust the next slice from the checkpointed cursor value
+        self._job_adjust_slice_from_checkpoint = True
+
+    def _reset_checkpointing(self) -> None:
+        # reseting the checkpoint flag, if bulk job has completed normally
+        self._job_adjust_slice_from_checkpoint = False
 
     def _job_completed(self) -> bool:
         return self._job_state == ShopifyBulkJobStatus.COMPLETED.value
@@ -183,8 +216,11 @@ class ShopifyBulkManager:
     def _job_canceled(self) -> bool:
         return self._job_state == ShopifyBulkJobStatus.CANCELED.value
 
+    def _job_failed(self) -> bool:
+        return self._job_state == ShopifyBulkJobStatus.FAILED.value
+
     def _job_cancel(self) -> None:
-        _, canceled_response = self._http_client.send_request(
+        _, canceled_response = self.http_client.send_request(
             http_method="POST",
             url=self.base_url,
             data=ShopifyBulkTemplates.cancel(self._job_id),
@@ -206,11 +242,14 @@ class ShopifyBulkManager:
             self._log_job_msg_count += 1
         else:
             message = f"Elapsed time: {self._job_elapsed_time_in_state} sec"
+            if self._job_last_rec_count > 0:
+                count_message = f". Rows collected: {self._job_last_rec_count}"
+                message = message + count_message
             self._log_state(message)
             self._log_job_msg_count = 0
 
     def _log_state(self, message: Optional[str] = None) -> None:
-        pattern = f"Stream: `{self.stream_name}`, the BULK Job: `{self._job_id}` is {self._job_state}"
+        pattern = f"Stream: `{self.http_client.name}`, the BULK Job: `{self._job_id}` is {self._job_state}"
         if message:
             self.logger.info(f"{pattern}. {message}.")
         else:
@@ -218,11 +257,14 @@ class ShopifyBulkManager:
 
     def _job_get_result(self, response: Optional[requests.Response] = None) -> Optional[str]:
         parsed_response = response.json().get("data", {}).get("node", {}) if response else None
-        job_result_url = parsed_response.get("url") if parsed_response and not self._job_self_canceled else None
+        # get `complete` or `partial` result from collected Bulk Job results
+        full_result_url = parsed_response.get("url") if parsed_response else None
+        partial_result_url = parsed_response.get("partialDataUrl") if parsed_response else None
+        job_result_url = full_result_url if full_result_url else partial_result_url
         if job_result_url:
             # save to local file using chunks to avoid OOM
             filename = self._tools.filename_from_url(job_result_url)
-            _, response = self._http_client.send_request(http_method="GET", url=job_result_url, request_kwargs={"stream": True})
+            _, response = self.http_client.send_request(http_method="GET", url=job_result_url, request_kwargs={"stream": True})
             response.raise_for_status()
             with open(filename, "wb") as file:
                 for chunk in response.iter_content(chunk_size=self._retrieve_chunk_size):
@@ -231,11 +273,24 @@ class ShopifyBulkManager:
                 file.write(END_OF_FILE.encode())
             return filename
 
+    def _job_get_checkpointed_result(self, response: Optional[requests.Response]) -> None:
+        if self._job_any_lines_collected or self._job_should_checkpoint:
+            # set the flag to adjust the next slice from the checkpointed cursor value
+            self._set_checkpointing()
+            # fetch the collected records from CANCELED Job on checkpointing
+            self._job_result_filename = self._job_get_result(response)
+
     def _job_update_state(self, response: Optional[requests.Response] = None) -> None:
         if response:
             self._job_state = response.json().get("data", {}).get("node", {}).get("status")
-            if self._job_state in [ShopifyBulkJobStatus.RUNNING.value, ShopifyBulkJobStatus.CANCELING.value]:
+            self._job_last_rec_count = int(response.json().get("data", {}).get("node", {}).get("objectCount", 0))
+
+            if self._job_state == ShopifyBulkJobStatus.RUNNING.value:
                 self._log_job_state_with_count()
+            elif self._job_state in [ShopifyBulkJobStatus.CANCELED.value, ShopifyBulkJobStatus.CANCELING.value]:
+                # do not emit `CANCELED / CANCELING` Bulk Job status, while checkpointing
+                if not self._job_should_checkpoint:
+                    self._log_job_state_with_count()
             else:
                 self._log_state()
 
@@ -245,19 +300,30 @@ class ShopifyBulkManager:
     def _on_canceled_job(self, response: requests.Response) -> Optional[AirbyteTracedException]:
         if not self._job_self_canceled:
             raise ShopifyBulkExceptions.BulkJobCanceled(
-                f"The BULK Job: `{self._job_id}` exited with {self._job_state}, details: {response.text}",
+                f"The BULK Job: `{self._job_id}` exited with {self._job_state}, details: {response.text}"
             )
+        else:
+            self._job_get_checkpointed_result(response)
 
     def _on_canceling_job(self, **kwargs) -> None:
         sleep(self._job_check_interval)
 
+    def _cancel_on_long_running_job(self) -> None:
+        self.logger.info(
+            f"Stream: `{self.http_client.name}` the BULK Job: {self._job_id} runs longer than expected ({self._job_max_elapsed_time} sec). Retry with the reduced `Slice Size` after self-cancelation."
+        )
+        self._job_cancel()
+
+    def _cancel_on_checkpointing(self) -> None:
+        self.logger.info(f"Stream: `{self.http_client.name}`, checkpointing after >= `{self._job_checkpoint_interval}` rows collected.")
+        # set the flag to adjust the next slice from the checkpointed cursor value
+        self._job_cancel()
+
     def _on_running_job(self, **kwargs) -> None:
         if self._is_long_running_job:
-            self.logger.info(
-                f"Stream: `{self.stream_name}` the BULK Job: {self._job_id} runs longer than expected ({self._job_max_elapsed_time} sec). Retry with the reduced `Slice Size` after self-cancelation."
-            )
-            # cancel the long-running bulk job
-            self._job_cancel()
+            self._cancel_on_long_running_job()
+        elif self._job_should_checkpoint:
+            self._cancel_on_checkpointing()
         else:
             sleep(self._job_check_interval)
 
@@ -265,9 +331,14 @@ class ShopifyBulkManager:
         self._job_result_filename = self._job_get_result(response)
 
     def _on_failed_job(self, response: requests.Response) -> AirbyteTracedException:
-        raise ShopifyBulkExceptions.BulkJobFailed(
-            f"The BULK Job: `{self._job_id}` exited with {self._job_state}, details: {response.text}",
-        )
+        if not self._supports_checkpointing:
+            raise ShopifyBulkExceptions.BulkJobFailed(
+                f"The BULK Job: `{self._job_id}` exited with {self._job_state}, details: {response.text}",
+            )
+        else:
+            # when the Bulk Job fails, usually there is a `partialDataUrl` available,
+            # we leverage the checkpointing in this case.
+            self._job_get_checkpointed_result(response)
 
     def _on_timeout_job(self, **kwargs) -> AirbyteTracedException:
         raise ShopifyBulkExceptions.BulkJobTimout(
@@ -283,14 +354,19 @@ class ShopifyBulkManager:
         raise ShopifyBulkExceptions.BulkJobError(f"Could not validate the status of the BULK Job `{self._job_id}`. Errors: {errors}.")
 
     def _on_non_handable_job_error(self, errors: List[Mapping[str, Any]]) -> AirbyteTracedException:
-        raise ShopifyBulkExceptions.BulkJobNonHandableError(f"The Stream: `{self.stream_name}`, Non-handable error occured: {errors}")
+        raise ShopifyBulkExceptions.BulkJobNonHandableError(f"The Stream: `{self.http_client.name}`, Non-handable error occured: {errors}")
 
-    def _collect_bulk_errors(self, response: requests.Response) -> List[Optional[dict]]:
+    def _get_server_errors(self, response: requests.Response) -> List[Optional[Mapping[str, Any]]]:
+        server_errors = response.json().get("errors", [])
+        return [server_errors] if isinstance(server_errors, str) else server_errors
+
+    def _get_user_errors(self, response: requests.Response) -> List[Optional[Mapping[str, Any]]]:
+        user_errors = response.json().get("data", {}).get("bulkOperationRunQuery", {}).get("userErrors", [])
+        return [user_errors] if isinstance(user_errors, str) else user_errors
+
+    def _collect_bulk_errors(self, response: requests.Response) -> List[Optional[Mapping[str, Any]]]:
         try:
-            server_errors = response.json().get("errors", [])
-            user_errors = response.json().get("data", {}).get("bulkOperationRunQuery", {}).get("userErrors", [])
-            errors = server_errors + user_errors
-            return errors
+            return self._get_server_errors(response) + self._get_user_errors(response)
         except (Exception, JSONDecodeError) as e:
             raise ShopifyBulkExceptions.BulkJobBadResponse(
                 f"Couldn't check the `response` for `errors`, status: {response.status_code}, response: `{response.text}`. Trace: {repr(e)}."
@@ -303,7 +379,7 @@ class ShopifyBulkManager:
             self._on_job_with_errors(errors)
 
     def _job_track_running(self) -> None:
-        _, response = self._http_client.send_request(
+        _, response = self.http_client.send_request(
             http_method="POST",
             url=self.base_url,
             data=ShopifyBulkTemplates.status(self._job_id),
@@ -311,7 +387,6 @@ class ShopifyBulkManager:
             request_kwargs={},
         )
         self._job_healthcheck(response)
-
         self._job_update_state(response)
         self._job_state_to_fn_map.get(self._job_state)(response=response)
 
@@ -339,11 +414,33 @@ class ShopifyBulkManager:
     def _has_reached_max_concurrency(self) -> bool:
         return self._concurrent_attempt == self._concurrent_max_retry
 
+    def _should_switch_shop_name(self, response: requests.Response) -> bool:
+        """
+        Sometimes the API returns the redirected response that points to the same Store but with different Name:
+        >> case:
+           -- The user inputs the `shop name` as "A":
+                while attempting to place the BULK Job
+           -- The response contains the redirected results to the `shop name` as "B", like:
+           response.url == "https://B.myshopify.com"
+
+        This redirection is related to:
+        1) `aliased` or `hidden` store names from being exposed
+        2) `migrated` store data to the `new store`, but referenced within the old one stil
+
+        reference issue: https://github.com/airbytehq/oncall/issues/5866
+        """
+        if self.base_url != response.url:
+            self.base_url = response.url
+            return True
+        return False
+
     @bulk_retry_on_exception(logger)
     def _job_check_state(self) -> None:
         while not self._job_completed():
             if self._job_canceled():
                 break
+            elif self._job_failed():
+                break
             else:
                 self._job_track_running()
 
@@ -354,7 +451,7 @@ class ShopifyBulkManager:
         else:
             query = self.query.get()
 
-        _, response = self._http_client.send_request(
+        _, response = self.http_client.send_request(
             http_method="POST",
             url=self.base_url,
             json={"query": ShopifyBulkTemplates.prepare(query)},
@@ -365,7 +462,10 @@ class ShopifyBulkManager:
         if self._has_running_concurrent_job(errors):
             # when the concurrent job takes place, another job could not be created
             # we typically need to wait and retry, but no longer than 10 min. (see retry in `bulk_retry_on_exception`)
-            raise ShopifyBulkExceptions.BulkJobCreationFailedConcurrentError(f"Failed to create job for stream {self.stream_name}")
+            raise ShopifyBulkExceptions.BulkJobCreationFailedConcurrentError(f"Failed to create job for stream {self.http_client.name}")
+        elif self._should_switch_shop_name(response):
+            # assign new shop name, since the one that specified in `config` was redirected to the different one.
+            raise ShopifyBulkExceptions.BulkJobRedirectToOtherShopError(f"Switching the `store` name, redirected to: {response.url}")
         else:
             # There were no concurrent error for this job so even if there were other errors, we can reset this
             self._concurrent_attempt = 0
@@ -384,7 +484,7 @@ class ShopifyBulkManager:
             self._job_id = bulk_response.get("id")
             self._job_created_at = bulk_response.get("createdAt")
             self._job_state = ShopifyBulkJobStatus.CREATED.value
-            self.logger.info(f"Stream: `{self.stream_name}`, the BULK Job: `{self._job_id}` is {ShopifyBulkJobStatus.CREATED.value}")
+            self.logger.info(f"Stream: `{self.http_client.name}`, the BULK Job: `{self._job_id}` is {ShopifyBulkJobStatus.CREATED.value}")
 
     def job_size_normalize(self, start: datetime, end: datetime) -> datetime:
         # adjust slice size when it's bigger than the loop point when it should end,
@@ -396,15 +496,43 @@ class ShopifyBulkManager:
         step = self._job_size if self._job_size else self._job_size_min
         return slice_start.add(days=step)
 
-    def get_adjusted_job_end(self, slice_start: datetime, slice_end: datetime) -> datetime:
+    def _adjust_slice_end(self, slice_end: datetime, checkpointed_cursor: Optional[str] = None) -> datetime:
+        """
+        Choose between the existing `slice_end` value or `checkpointed_cursor` value, if provided.
+        """
+        return pdm.parse(checkpointed_cursor) if checkpointed_cursor else slice_end
+
+    def get_adjusted_job_end(self, slice_start: datetime, slice_end: datetime, checkpointed_cursor: Optional[str] = None) -> datetime:
+        if self._job_adjust_slice_from_checkpoint:
+            # set the checkpointing to default, before the next slice is emitted, to avoid inf.loop
+            self._reset_checkpointing()
+            return self._adjust_slice_end(slice_end, checkpointed_cursor)
+
         if self._is_long_running_job:
             self._job_size_reduce_next()
             return slice_start
+
+        return slice_end
+
+    def _emit_final_job_message(self, job_current_elapsed_time: int) -> None:
+        final_message = f"Stream: `{self.http_client.name}`, the BULK Job: `{self._job_id}` time elapsed: {job_current_elapsed_time} sec."
+
+        if self._job_any_lines_collected:
+            lines_collected_message = f" Rows collected: {self._job_last_rec_count} --> records: `{self.record_producer.record_composed}`."
+            final_message = final_message + lines_collected_message
+
+        # emit final Bulk job status message
+        self.logger.info(f"{final_message}")
+
+    def _process_bulk_results(self) -> Iterable[Mapping[str, Any]]:
+        if self._job_result_filename:
+            # produce records from saved bulk job result
+            yield from self.record_producer.read_file(self._job_result_filename)
         else:
-            return slice_end
+            yield from []
 
     @limiter.balance_rate_limit(api_type=ApiTypeEnum.graphql.value)
-    def job_check_for_completion(self) -> Optional[str]:
+    def job_get_results(self) -> Optional[Iterable[Mapping[str, Any]]]:
         """
         This method checks the status for the `CREATED` Shopify BULK Job, using it's `ID`.
         The time spent for the Job execution is tracked to understand the effort.
@@ -414,7 +542,7 @@ class ShopifyBulkManager:
         try:
             # track created job until it's COMPLETED
             self._job_check_state()
-            return self._job_result_filename
+            yield from self._process_bulk_results()
         except (
             ShopifyBulkExceptions.BulkJobFailed,
             ShopifyBulkExceptions.BulkJobTimout,
@@ -426,7 +554,8 @@ class ShopifyBulkManager:
             raise bulk_job_error
         finally:
             job_current_elapsed_time = round((time() - job_started), 3)
-            self.logger.info(f"Stream: `{self.stream_name}`, the BULK Job: `{self._job_id}` time elapsed: {job_current_elapsed_time} sec.")
+            # emit the final Bulk Job log message
+            self._emit_final_job_message(job_current_elapsed_time)
             # check whether or not we should expand or reduce the size of the slice
             self.__adjust_job_size(job_current_elapsed_time)
             # reset the state for COMPLETED job
