@@ -24,25 +24,20 @@ def bulk_retry_on_exception(logger: logging.Logger, more_exceptions: Optional[Tu
     def decorator(func: Callable) -> Callable:
         @wraps(func)
         def wrapper(self, *args, **kwargs) -> Any:
-            # mandatory class attributes
-            max_retries = self._job_max_retries
-            stream_name = self.stream_name
-            backoff_time = self._job_backoff_time
-
             current_retries = 0
             while True:
                 try:
                     return func(self, *args, **kwargs)
                 except BULK_RETRY_ERRORS or more_exceptions as ex:
                     current_retries += 1
-                    if current_retries > max_retries:
+                    if current_retries > self._job_max_retries:
                         logger.error("Exceeded retry limit. Giving up.")
                         raise
                     else:
                         logger.warning(
-                            f"Stream `{stream_name}`: {ex}. Retrying {current_retries}/{max_retries} after {backoff_time} seconds."
+                            f"Stream `{self.http_client.name}`: {ex}. Retrying {current_retries}/{self._job_max_retries} after {self._job_backoff_time} seconds."
                         )
-                        sleep(backoff_time)
+                        sleep(self._job_backoff_time)
                 except ShopifyBulkExceptions.BulkJobCreationFailedConcurrentError:
                     if self._concurrent_attempt == self._concurrent_max_retry:
                         message = f"The BULK Job couldn't be created at this time, since another job is running."
@@ -51,9 +46,13 @@ def bulk_retry_on_exception(logger: logging.Logger, more_exceptions: Optional[Tu
 
                     self._concurrent_attempt += 1
                     logger.warning(
-                        f"Stream: `{self.stream_name}`, the BULK concurrency limit has reached. Waiting {self._concurrent_interval} sec before retry, attempt: {self._concurrent_attempt}.",
+                        f"Stream: `{self.http_client.name}`, the BULK concurrency limit has reached. Waiting {self._concurrent_interval} sec before retry, attempt: {self._concurrent_attempt}.",
                     )
                     sleep(self._concurrent_interval)
+                except ShopifyBulkExceptions.BulkJobRedirectToOtherShopError:
+                    logger.warning(
+                        f"Stream: `{self.http_client.name}`, the `shop name` differs from the provided in `input configuration`. Switching to the `{self._tools.shop_name_from_url(self.base_url)}`.",
+                    )
 
         return wrapper
 
